import gzip
import os
import subprocess


class KaldiPipe(object):
    def __init__(self, process, *_, rxtype):
        self.wait = process.wait
        self.out_stream = process.stdout
        self.rxtype = rxtype

    def close(self):
        self.wait()

    @property
    def bytes(self):
        if not hasattr(self, "_bytes"):
            self._bytes = self.out_stream.read()
        return self._bytes


class KaldiFile(object):
    def __init__(self, filename, *_, rxtype, is_gz=False):
        self.filename = filename
        self.is_gz = is_gz
        self.rxtype = rxtype

    def __getstate__(self):
        # Add file stats to pickle for timestamp uniqueness
        state = dict(self.__dict__)
        state['__filestate'] = (os.path.getctime(self.filename), os.path.getmtime(self.filename))
        return state


class KaldiGzFile(KaldiFile):
    def __init__(self, filename, *_, rxtype):
        super().__init__(filename, rxtype=rxtype, is_gz=True)


class KaldiBytes(object):
    def __init__(self, bytes, *_, rxtype):
        self.bytes = bytes
        self.rxtype = rxtype


class KaldiBytesArk(KaldiBytes):
    def __init__(self, dict):
        super().__init__(dict, rxtype='ark')

    @property
    def bytes(self):
        return b''.join([
            id.encode() + b' ' + b for id, b in self._bytes.items()
        ])

    @bytes.setter
    def bytes(self, value):
        self._bytes = value



import kaldibin


def lattice_1best(
        lattice,
        *_,
        acoustic_scale=1.0,
        lm_scale=1.0,
        word_ins_penalty=0.0
):
    return kaldibin._context.run(
        "src/latbin/lattice-1best",
        "--acoustic-scale={}".format(acoustic_scale),
        "--lm-scale={}".format(lm_scale),
        "--word-ins-penalty={}".format(word_ins_penalty),
        lattice,
        wxtype="ark",
        wxfilename="-"
    )


def lattice_align_phones(
        model,
        lattice,
        output_error_lats=True,
        remove_epsilon=True,
        reorder=True,
        replace_output_symbols=False
):
    return kaldibin._context.run(
        "src/latbin/lattice-align-phones",
        "--output-error-lats={}".format("true" if output_error_lats else "false"),
        "--remove-epsilon={}".format("true" if remove_epsilon else "false"),
        "--reorder={}".format("true" if reorder else "false"),
        "--replace-output-symbols={}".format("true" if replace_output_symbols else "false"),
        model,
        lattice,
        wxtype="ark",
        wxfilename = "-"
    )


def lattice_align_words_lexicon(
        lexicon,
        model,
        lattice,
        *_,
        allow_duplicate_paths=False,
        max_expand=-1,
        output_error_lats=True,
        output_if_empty=False,
        partial_word_label=0,
        reorder=True,
        test=False
):
    return kaldibin._context.run(
        "src/latbin/lattice-align-words-lexicon",
        "--allow-duplicate-paths={}".format("true" if allow_duplicate_paths else "false"),
        "--max-expand={}".format(max_expand),
        "--output-error-lats={}".format("true" if output_error_lats else "false"),
        "--output-if-empty={}".format("true" if output_if_empty else "false"),
        "--partial-word-label={}".format(partial_word_label),
        "--reorder={}".format("true" if reorder else "false"),
        lexicon,
        model,
        lattice,
        wxtype = "ark",
        wxfilename = "-"
    )


def lattice_determinize(
        lattice,
        *_,
        acoustic_sale=1.0,
        beam=10.0,
        beam_ratio=0.9,
        delta=0.000976562,
        max_loop=500000,
        max_mem=50000000,
        minimize=False,
        num_loops=20,
        prune=False
):
    return kaldibin._context.run(
        "src/latbin/lattice-determinize",
        "--acoustic-scale={}".format(acoustic_sale),
        "--beam={}".format(beam),
        "--beam-ratio={}".format(beam_ratio),
        "--delta={}".format(delta),
        "--max-loop={}".format(max_loop),
        "--max-mem={}".format(max_mem),
        "--minimize={}".format("true" if minimize else "false"),
        "--num-loops={}".format(num_loops),
        "--prune={}".format("true" if prune else "false"),
        lattice,
        wxtype="ark",
        wxfilename="-"
    )


def lattice_to_ctm_conf(
        lattice,
        *_,
        acoustic_scale=1.0,
        confidence_digits=2,
        decode_mbr=True,
        frame_shift=0.01,
        inv_acoustic_scale=1.0,
        lm_scale=1.0,
        print_silence=False
):
    return kaldibin._context.run(
        "src/latbin/lattice-to-ctm-conf",
        "--acoustic-scale={}".format(float(acoustic_scale)),
        "--confidence-digits={}".format(int(confidence_digits)),
        "--decode-mbr={}".format("true" if decode_mbr else "false"),
        "--frame-shift={}".format(frame_shift),
        "--inv-acoustic-scale={}".format(inv_acoustic_scale),
        "--lm-scale={}".format(lm_scale),
        "--print-silence={}".format("true" if print_silence else "false"),
        lattice,
        wxtype = None,
        wxfilename = "-"
    )


def nbest_to_ctm(
        nbest,
        *_,
        frame_shift=0.01,
        precision=2,
        print_silence=False
):
    return kaldibin._context.run(
        "src/latbin/nbest-to-ctm",
        "--frame-shift={}".format(float(frame_shift)),
        "--precision={}".format(int(precision)),
        "--print-silence={}".format("true" if print_silence else "false"),
        nbest,
        wxtype = None,
        wxfilename = "-"
    )



import os
import subprocess
import tempfile
import threading
import uuid

import kaldibin
from kaldibin.adapters.io import KaldiPipe, KaldiFile, KaldiBytes

_KALDI_ROOT = os.environ['KALDI_ROOT'] if 'KALDI_ROOT' in os.environ else None


class KaldiContext(object):
    def __init__(self, kaldi_root=_KALDI_ROOT):
        if not kaldi_root or not os.path.isdir(kaldi_root):
            raise Exception('Kaldi root not found. Set the environment variable KALDI_ROOT.')
        self._to_close = []
        self._to_delete = []
        self._kaldi_root = kaldi_root

    def open(self):
        self._previous_context = kaldibin._context
        kaldibin._context = self
        return self

    def close(self):
        for to_close in self._to_close:
            to_close.close()
        for to_delete in self._to_delete:
            os.unlink(to_delete)
        kaldibin._context = self._previous_context

    def run(self, executable, *args, input=None, wxtype, wxfilename):
        exe = os.path.join(self._kaldi_root, executable)
        prepared_args, pipe_handles = _prepare_args(self, args)
        wxspecifier = '{}:{}'.format(wxtype, wxfilename) if wxtype else wxfilename
        process = subprocess.Popen(
            [exe, *prepared_args, wxspecifier],
            stdin=subprocess.PIPE if input is not None else None,
            stdout=subprocess.PIPE,
            stderr=None,
            close_fds=True,
        )
        if input is not None:
            process.communicate(input=input)

        # Run FIFO inputs as separate thread to avoid deadlock
        fifo_args = [arg for arg in args if _is_fifo(arg)]
        for arg, fifo in zip(fifo_args, pipe_handles):
            threading.Thread(target=_write_fifo, args=(arg, fifo)).start()

        if wxfilename == '-':
            pipe_out = KaldiPipe(process, rxtype=wxtype)
            self._to_close.append(pipe_out)
            return pipe_out
        else:
            process.wait()
            return KaldiFile(wxfilename, rxtype=wxtype)


def _prepare_args(kaldi_context, args):
    prepared = []
    fifos = []
    for arg in args:
        if _is_fifo(arg):
            fifo_filename = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
            os.mkfifo(fifo_filename)
            kaldi_context._to_delete.append(fifo_filename)
            fifos.append(fifo_filename)
            prepared.append('{}:{}'.format(arg.rxtype, fifo_filename) if arg.rxtype else fifo_filename)
        elif _is_rwspecifier(arg):
            if hasattr(arg, "is_gz") and arg.is_gz:
                prepared.append('{}:gunzip -c {}|'.format(arg.rxtype, arg.filename))
            elif arg.rxtype:
                prepared.append('{}:{}'.format(arg.rxtype, arg.filename) if arg.rxtype else arg.filename)
            else:
                prepared.append(arg.filename)
        else:
            prepared.append(str(arg))
    return prepared, fifos


def _write_fifo(arg, fifo):
    write_handle = os.open(fifo, os.O_WRONLY | os.O_NOCTTY)
    if type(arg) is KaldiBytes or issubclass(type(arg), KaldiBytes):
        os.write(write_handle, arg.bytes)
    else:
        subprocess.run(['cat'], stdin=arg.out_stream, stdout=write_handle, shell=False, close_fds=True)
    os.close(write_handle)


def _is_fifo(arg):
    if type(arg) is KaldiBytes or issubclass(type(arg), KaldiBytes):
        return True
    return hasattr(arg, 'rxtype') and \
           hasattr(arg, 'out_stream') and \
           arg.out_stream is not None


def _is_rwspecifier(arg):
    return hasattr(arg, 'rxtype') and hasattr(arg, 'filename')



import atexit
import os

from kaldibin.adapters.io import KaldiBytes, KaldiFile, KaldiGzFile, KaldiPipe, KaldiBytesArk
from kaldibin.adapters.context import KaldiContext
from kaldibin.featbin import compute_cmvn_stats, compute_mfcc_feats, wav_copy
from kaldibin.latbin import lattice_1best, lattice_align_phones, lattice_align_words_lexicon, lattice_determinize,\
    lattice_to_ctm_conf, nbest_to_ctm
from kaldibin.online2bin import ivector_extract_online2, online2_wav_nnet3_latgen_faster

_KALDI_ROOT = os.environ['KALDI_ROOT'] if 'KALDI_ROOT' in os.environ else None
_context = KaldiContext(_KALDI_ROOT)
_context.open()
atexit.register(_context.close)



import kaldibin


def nnet3_latgen_faster(
        nnet,
        fst,
        features,
        *_,
        acoustic_scale=0.1,
        allow_partial=False,
        beam=16,
        beam_delta=0.5,
        computation_debug=False,
        debug_computation=False,
        delta=0.000976562,
        determinize_lattice=True,
        extra_left_context=0,
        extra_left_context_initial=-1,
        extra_right_context=0,
        extra_right_context_final=-1,
        frame_subsampling_factor=1,
        frames_per_chunk=50,
        hash_ratio=2.0,
        ivectors='',
        lattice_beam=10.0,
        max_active=2147483647,
        max_mem=50000000,
        min_active=200,
        minimize=False,
        online_ivector_period=0,
        online_ivectors='',
        phone_determinize=True,
        prune_interval=25,
        utt2spk='',
        word_determinize=True,
        word_symbol_table='',
        **args
    ):
    '''
    WARNING: INCOMPLETE WRAPPER IMPLEMENTATION

    :param nnet:
    :param fst:
    :param features:
    :param _:
    :param acoustic_scale:
    :param allow_partial:
    :param beam:
    :param beam_delta:
    :param computation_debug:
    :param debug_computation:
    :param delta:
    :param determinize_lattice:
    :param extra_left_context:
    :param extra_left_context_initial:
    :param extra_right_context:
    :param extra_right_context_final:
    :param frame_subsampling_factor:
    :param frames_per_chunk:
    :param hash_ratio:
    :param ivectors:
    :param lattice_beam:
    :param max_active:
    :param max_mem:
    :param min_active:
    :param minimize:
    :param online_ivector_period:
    :param online_ivectors:
    :param phone_determinize:
    :param prune_interval:
    :param utt2spk:
    :param word_determinize:
    :param word_symbol_table:
    :param args:
    :return:
    '''

    return kaldibin._context.run(
        'src/nnet3bin/nnet3-latgen-faster',
        '--frames-per-chunk={}'.format(frames_per_chunk),
        '--extra-left-context={}'.format(extra_left_context),
        '--extra-right-context={}'.format(extra_right_context),
        '--extra-left-context-initial={}'.format(extra_left_context_initial),
        '--extra-right-context-final={}'.format(extra_right_context_final),
        '--minimize={}'.format('true' if minimize else 'false'),
        '--max-active={}'.format(max_active),
        '--max-mem={}'.format(max_mem),
        '--min-active={}'.format(min_active),
        '--beam={}'.format(beam),
        '--lattice-beam={}'.format(lattice_beam),
        '--acoustic-scale={}'.format(acoustic_scale),
        '--allow-partial={}'.format('true' if allow_partial else 'false'),
        '--word-symbol-table={}'.format(word_symbol_table),
        '--frame-subsampling-factor={}'.format(frame_subsampling_factor),
        '--online-ivectors={}'.format(online_ivectors),
        '--online-ivector-period={}'.format(online_ivector_period),
        nnet,
        fst,
        features,
        wxtype="ark",
        wxfilename="-"
    )



import kaldibin


def compute_cmvn_stats(
        feats,
        *_,
        binary=True,
        spk2utt='',
        weights='',
        wxtype='ark',
        wxfilename='-'
):
    return kaldibin._context.run(
        'src/featbin/compute-cmvn-stats',
        '--binary={}'.format('true' if binary else 'false'),
        '--spk2utt={}'.format(spk2utt),
        '--weights={}'.format(weights),
        feats,
        wxtype=wxtype,
        wxfilename=wxfilename
    )


def compute_mfcc_feats(
        wavs,
        *_,
        allow_downsample=False,
        blackman_coeff=0.42,
        cepstral_lifter=22,
        channel=-1,
        debug_mel=False,
        dither=1.0,
        energy_floor=0.0,
        frame_length=25,
        frame_shift=10,
        high_freq=0,
        htk_compat=False,
        low_freq=20,
        min_duration=0,
        num_ceps=13,
        num_mel_bins=23,
        output_format='kaldi',
        preemphasis_coefficient=0.97,
        raw_energy=True,
        remove_dc_offset=True,
        round_to_power_of_two=True,
        sample_frequency=16000,
        snip_edges=True,
        subtract_mean=False,
        use_energy=True,
        utt2spk='',
        vtln_high=-500,
        vtln_low=100,
        vtln_map='',
        vtln_warp=1,
        window_type='povey',
        verbose=0,
        config='',
        wxtype='ark',
        wxfilename='-'
):
    return kaldibin._context.run(
        'src/featbin/compute-mfcc-feats',
        '--allow-downsample={}'.format('true' if allow_downsample else 'false'),
        '--blackman-coeff={}'.format(blackman_coeff),
        '--cepstral-lifter={}'.format(cepstral_lifter),
        '--channel={}'.format(channel),
        '--debug-mel={}'.format('true' if debug_mel else 'false'),
        '--dither={}'.format(dither),
        '--energy-floor={}'.format(energy_floor),
        '--frame-length={}'.format(frame_length),
        '--frame-shift={}'.format(frame_shift),
        '--high-freq={}'.format(high_freq),
        '--htk-compat={}'.format('true' if htk_compat else 'false'),
        '--low-freq={}'.format(low_freq),
        '--min-duration={}'.format(min_duration),
        '--num-ceps={}'.format(num_ceps),
        '--num-mel-bins={}'.format(num_mel_bins),
        '--output-format={}'.format(output_format),
        '--preemphasis_coefficient={}'.format(preemphasis_coefficient),
        '--raw-energy={}'.format('true' if raw_energy else 'false'),
        '--remove-dc-offset={}'.format('true' if remove_dc_offset else 'false'),
        '--round-to-power-of-two={}'.format('true' if round_to_power_of_two else 'false'),
        '--sample-frequency={}'.format(sample_frequency),
        '--snip-edges={}'.format('true' if snip_edges else 'false'),
        '--subtract-mean={}'.format('true' if subtract_mean else 'false'),
        '--use-energy={}'.format('true' if use_energy else 'false'),
        '--utt2spk={}'.format(utt2spk),
        '--vtln-high={}'.format(vtln_high),
        '--vtln-low={}'.format(vtln_low),
        '--vtln-map={}'.format(vtln_map),
        '--vtln-warp={}'.format(vtln_warp),
        '--window-type={}'.format(window_type),
        '--verbose={}'.format(verbose),
        '--config={}'.format(config),
        wavs,
        wxtype=wxtype,
        wxfilename=wxfilename
    )


def wav_copy(wav, *_, wxtype='ark', wxfilename='-'):
    return kaldibin._context.run(
        'src/featbin/wav-copy',
        wav,
        wxtype=wxtype,
        wxfilename=wxfilename
    )


import setuptools

# with open("README.md", "r") as fh:
#     long_description = fh.read()

setuptools.setup(
    name='kaldibin',
    version='0.4.0',
    author='Robert Gale',
    author_email='rcgale@gmail.com',
    packages=[
        'kaldibin',
        'kaldibin.adapters',
    ],
    url='https://github.com/rcgale/kaldibin',
    description='Python wrappers for Kaldi executables. Still in a partially-implemented, proof-of-concept stage.',
    install_requires=[],
)




import kaldibin


def ivector_extract_online2(
        spk2utt,
        feats,
        *_,
        config='',
        wxtype='ark',
        wxfilename='-'
):
    '''
    WARNING: INCOMPLETE WRAPPER IMPLEMENTATION
    :param spk2utt:
    :param feats:
    :param _:
    :param config:
    :param wxtype:
    :param wxfilename:
    :return:
    '''

    return kaldibin._context.run(
        'src/online2bin/ivector-extract-online2',
        '--config={}'.format(config),
        spk2utt,
        feats,
        wxtype=wxtype,
        wxfilename=wxfilename
    )


def online2_wav_nnet3_latgen_faster(
        nnet3,
        fst,
        spk2utt,
        wav,
        *_,
        max_active=2147483647,
        beam=16.0,
        do_endpointing=False,
        frame_subsampling_factor=1,
        lattice_beam=10.0,
        acoustic_scale=0.1,
        word_symbol_table='',
        config='',
        wxtype='ark',
        wxfilename='-'
):
    '''
    WARNING: INCOMPLETE WRAPPER IMPLEMENTATION
    :param nnet3:
    :param fst:
    :param spk2utt:
    :param wav:
    :param _:
    :param max_active:
    :param beam:
    :param do_endpointing:
    :param frame_subsampling_factor:
    :param lattice_beam:
    :param acoustic_scale:
    :param word_symbol_table:
    :param config:
    :param wxtype:
    :param wxfilename:
    :return:
    '''

    return kaldibin._context.run(
        "src/online2bin/online2-wav-nnet3-latgen-faster",
        "--online=false",
        "--do-endpointing={}".format('true' if do_endpointing else 'false'),
        "--config={}".format(config),
        "--frame-subsampling-factor={}".format(frame_subsampling_factor),
        "--max-active={}".format(max_active),
        "--beam={}".format(beam),
        "--lattice-beam={}".format(lattice_beam),
        "--acoustic-scale={}".format(acoustic_scale),
        "--word-symbol-table={}".format(word_symbol_table),
        nnet3,
        fst,
        spk2utt,
        wav,
        wxtype=wxtype,
        wxfilename=wxfilename
    )